# -*- coding: utf-8 -*-
'''
This module contains unit tests for functions and classes provided by
spam_lists.structures module
'''
from __future__ import unicode_literals

# pylint: disable=redefined-builtin
from builtins import str, range, object
from nose_parameterized import parameterized

from spam_lists.exceptions import InvalidHostError, InvalidHostnameError
from spam_lists.structures import (
    Hostname, create_host, IPv4Address, IPv6Address
)
from test.compat import unittest, Mock, patch, MagicMock


class BaseHostTest(object):
    ''' A class providing tests for subclasses of Host class

    :var tested_instance: an instance of tested class to be used
    in tests
    :var class_to_test: a class to be tested
    '''

    def setUp(self):
        self.tested_instance.value = MagicMock()

    def test_lt_for_smaller_value(self):
        self.tested_instance.value.__lt__.return_value = False
        self.assertFalse(self.tested_instance < Mock())

    def test_lt_for_larger_value(self):
        self.tested_instance.value.__lt__.return_value = True
        self.assertTrue(self.tested_instance < Mock())

    def test_lt_for_other_not_having_value_attribute(self):
        other = Mock(spec=[])
        self.assertEqual(
            NotImplemented,
            self.tested_instance.__lt__(other)
        )

    def test_lt_for_other_not_having_to_unicode_method(self):
        other = Mock(spec=['value'])
        value = MagicMock()
        value.__lt__.side_effect = TypeError
        self.tested_instance.value = value
        self.assertEqual(
            NotImplemented,
            self.tested_instance.__lt__(other)
        )


class HostnameTest(BaseHostTest, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for Hostname class

    :var superdomain_str: a string value representing a parent of
     a domain used to create tested instance
    :var domain_str: a string value used as an argument to
    constructor when creating tested instance
    :var subdomain_str: a string value representing a child domain
    to the one used to create tested instance
    :var superdomain: a Hostname instance representing a parent
    of the tested instance
    :var domain: a Hostname instance to be tested
    :var subdomain: a Hostname instance representing a child
    of the tested instance
    :var unrelated_domain: a Hostname instance representing
    a domain unrelated to the one represented by tested instance
    '''
    class_to_test = Hostname
    superdomain_str = 'superdomain.com'
    domain_str = 'domain.'+superdomain_str
    subdomain_str = 'subdomain.'+domain_str
    superdomain = class_to_test(superdomain_str)
    domain = class_to_test(domain_str)
    subdomain = class_to_test(subdomain_str)
    unrelated_domain = class_to_test('other.com')
    tested_instance = class_to_test('compared.com')

    @parameterized.expand([
        ('hostname', '-e'),
        ('hostname', '/e'),
        ('argument', 123)
    ])
    def test_constructor_for_invalid(self, _, value):
        self.assertRaises(InvalidHostnameError, Hostname, value)

    @parameterized.expand([
        ('unrelated_domain', unrelated_domain, False),
        ('a_subdomain', subdomain, False),
        ('non_hostname_object', '123.4.5.11', False),
        ('the_same_domain', domain, True),
        ('a_superdomain', superdomain, True)
    ])
    def test_is_subdomain_for(self, _, other, expected):
        actual = self.domain.is_subdomain(other)
        if expected:
            self.assertTrue(actual)
        else:
            self.assertFalse(actual)

    @parameterized.expand([
        ('returns_false', False),
        ('returns_true', True)
    ])
    def test_lt_for_not_comparable_values(self, _, result):
        self.tested_instance.value.__lt__.side_effect = TypeError

        str_value = self.tested_instance.to_unicode()
        str_value.__lt__.return_value = result
        other = Mock()

        assertion = self.assertTrue if result else self.assertFalse
        assertion(self.tested_instance < other)


class IPAddressTestMixin(BaseHostTest):
    ''' A class providing tests for subclasses of IPAddress

    :var class_to_test: a subclass of IPAddress to be tested
    '''
    def setUp(self):
        self.value_constructor_patcher = patch.object(
            self.class_to_test,
            'factory',
            Mock()
        )
        self.value_constructor_mock = self.value_constructor_patcher.start()

        self.name_from_ip_patcher = patch('spam_lists.structures.name_from_ip')
        self.name_from_ip_mock = self.name_from_ip_patcher.start()

        self.tested_instance = self.class_to_test(Mock())
        super(IPAddressTestMixin, self).setUp()

    def tearDown(self):
        self.value_constructor_patcher.stop()
        self.name_from_ip_patcher.stop()

    def test_constructor_for_invalid_argument(self):
        self.value_constructor_mock.side_effect = ValueError
        self.assertRaises(
            self.class_to_test.invalid_ip_error_type,
            self.class_to_test,
            Mock()
        )

    def test_create_relative_domain_for_ip(self):
        self.tested_instance.relative_domain

        self.name_from_ip_mock.assert_called_once_with(
            str(self.tested_instance.value)
        )
        name = self.name_from_ip_mock.return_value
        name.relativize.assert_called_once_with(
            self.class_to_test.reverse_domain
        )

    def test_relative_domain_value(self):
        name = self.name_from_ip_mock.return_value
        expected = name.relativize.return_value
        actual = self.tested_instance.relative_domain
        self.assertEqual(expected, actual)

    def test_lt_for_not_comparable_values(self):
        self.tested_instance.value.__lt__.side_effect = TypeError

        str_value = self.tested_instance.to_unicode()
        other = Mock()
        other_str_value = 'other_str'
        other.to_unicode.return_value = other_str_value

        self.assertEqual(
            str_value < other_str_value,
            self.tested_instance < other
        )


class IPv4AddressTest(IPAddressTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    class_to_test = IPv4Address


class IPv6AddressTest(IPAddressTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    class_to_test = IPv6Address


class CreateHostTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for create_host function

    :var factories: a list of mocks representing factories used by
    the function during tests
    '''
    def setUp(self):
        self.factories = tuple(Mock() for _ in range(5))

    @parameterized.expand([
        ('v4', '127.0.0.1'),
        ('v6', '2001:db8:abc:125::45'),
    ])
    def test_host_for_ip(self, _, value):
        ip_address = self.factories[0]
        expected = ip_address(value)
        actual = create_host(self.factories, value)
        self.assertEqual(actual, expected)

    def test_host_for_hostname(self):
        ''' The function is expected to return an object
        returned by a host factory for given hostname'''
        for i, factory in enumerate(self.factories):
            if i != 1:
                factory.side_effect = InvalidHostError
        host_factory = self.factories[1]
        host_value = 'abc.com'
        expected = host_factory(host_value)
        actual = create_host(self.factories, host_value)
        self.assertEqual(expected, actual)

    @parameterized.expand([
        ('ipv4', '299.0.0.1'),
        ('ipv4', '99.22.33.1.23'),
        ('ipv6', '2001:db8:abc:125::4h'),
        ('ipv6', '2001:db8:abcef:125::43'),
        ('hostname', '-e'),
        ('hostname', '/e')
    ])
    def test_host_for_invalid(self, _, value):
        for factory in self.factories:
            factory.side_effect = InvalidHostError
        self.assertRaises(InvalidHostError, create_host, self.factories, value)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
'''
A common module for compatibility related imports and
definitions used during testing
'''
# pylint: disable=unused-import

from __future__ import unicode_literals

import unittest

try:
    from unittest.mock import Mock, MagicMock, patch  # @NoMove
except ImportError:
    # pylint: disable=import-error
    from mock import Mock, MagicMock, patch  # @NoMove @UnusedImport

from six import assertCountEqual, PY2

from spam_lists.compat import lru_cache  # @NoMove @UnusedImport


class Py2TestCase(unittest.TestCase):
    def assertCountEqual(self, expected_sequence, actual_sequence, msg=None):
        return assertCountEqual(self, expected_sequence, actual_sequence, msg)


if PY2:
    unittest.TestCase = Py2TestCase



# -*- coding: utf-8 -*-

'''
This module contains classes of clients of online services that
can be queried to check if a given hostname, IP address or URL
is recognized as spam.

It also contains instances of those of the classes that represent clients
of services that can be ready for use without providing custom, user-specific
data, like API codes or application identifiers.
'''
from __future__ import unicode_literals

# pylint: disable=redefined-builtin
from builtins import zip, str, range, object
from dns import name
from dns.resolver import NXDOMAIN, query
from future.utils import raise_from
from requests import get, post
from requests.exceptions import HTTPError

from .exceptions import UnathorizedAPIKeyError, UnknownCodeError
from .host_list import HostList
from .structures import (
    AddressListItem, non_ipv6_host, ip_address, registered_domain,
    registered_domain_or_ip
)

from .validation import accepts_valid_urls


class DNSBL(HostList):
    ''' Represents a DNSBL service '''
    def __init__(
            self,
            identifier,
            query_suffix,
            classification_map,
            host_factory):
        ''' Create new DNSBL object

        :param identifier: a value designating DNSBL service provider:
        its name or url address.
        :param query_suffix: a suffix added to DNSBL query address
        :param classification_map: item classes associated with
        DNSBL query return codes
        :param host_factory: a callable object that returns an object
         representing host and providing method for getting a relative
         domain pertaining to it.
        '''
        self._identifier = identifier
        self._query_suffix = name.from_text(query_suffix)
        self._classification_map = classification_map
        self._host_factory = host_factory
        super(DNSBL, self).__init__(host_factory)

    def _query(self, host_object):
        ''' Query DNSBL service for given value

        :param host_object: an object representing host,
         created by _host_factory
        :returns: an instance of dns.resolver.Answer for
         given value, if it is listed. Otherwise, it returns None
        '''
        host_to_query = host_object.relative_domain
        query_name = host_to_query.derelativize(self._query_suffix)
        try:
            return query(query_name)
        except NXDOMAIN:
            return None

    def __str__(self):
        return str(self._identifier)

    def _contains(self, host_object):
        return bool(self._query(host_object))

    def _get_entry_classification(self, code):
        return [self._classification_map[code]]

    def _get_match_and_classification(self, host_object):
        answers = self._query(host_object)
        if answers is None:
            return None, None
        try:
            classification = set()
            for answer in answers:
                last_octet = answer.to_text().split('.')[-1]
                classes = self._get_entry_classification(int(last_octet))
                classification.update(classes)
            return host_object, classification
        except KeyError as ex:
            msg_tpl = "The code '{}' has no corresponding classification value"
            msg = msg_tpl.format(ex.args[0])
            raise_from(UnknownCodeError(msg), ex)


def get_powers_of_2(_sum):
    ''' Get powers of with a given sum

    This function transforms given integer to a binary string.
    A reversed value limited to digits of binary number is extracted
    from it, and each of its characters is enumerated.

    Each digit is tested for not being 0. If the test passes, the index
    associated with the digit is used as an exponent to get the next
    value in the sequence to be returned.

    :param _sum: a sum of all elements of the sequence to be returned
    :returns: a list of powers of two whose sum is given
    '''
    return [2**y for y, x in enumerate(bin(_sum)[:1:-1]) if int(x)]


class BitmaskingDNSBL(DNSBL):
    ''' A class representing DNSBL services mapping listed items to
    sums of classification codes

    Each classification code is a power of two.

    '''
    def _get_entry_classification(self, code):
        codes = get_powers_of_2(code)
        return [cl for c in codes for cl
                in DNSBL._get_entry_classification(self, c)]


class HpHosts(HostList):
    ''' hpHosts client '''
    identifier = ' http://www.hosts-file.net/'
    _NOT_LISTED = 'Not Listed'

    def __init__(self, client_name):
        '''Constructor

        :param client_name: name of client using the service
        '''
        self.app_id = client_name
        super(HpHosts, self).__init__(non_ipv6_host)

    def _query(self, host_object, classification=False):
        ''' Query the client for data of given host

        :param host_object: an object representing a host value
        :param classification: if True: hpHosts is queried also
         for classification for given host, if listed
        :returns: content of response to GET request to hpHosts
         for data on the given host
        '''
        template = 'http://verify.hosts-file.net/?v={}&s={}'
        url = template.format(self.app_id, host_object.to_unicode())
        url = url + '&class=true' if classification else url
        return get(url).text

    def _contains(self, host_object):
        return self._NOT_LISTED not in self._query(host_object)

    def _get_match_and_classification(self, host_object):
        data = self._query(host_object, True)
        if self._NOT_LISTED in data:
            return None, None
        elements = data.split(',')
        classification = set(elements[1:])
        return host_object, classification


class GoogleSafeBrowsing(object):
    ''' Google Safe Browsing lookup API client '''
    protocol_version = '3.1'
    max_urls_per_request = 500

    def __init__(self, client_name, app_version, api_key):
        ''' Create new instance

        :param client_name: name of application using the API
        :param app_version: version of the application
        :param api_key: API key given by Google:
        https://developers.google.com/safe-browsing/key_signup
        '''
        self.api_key = api_key
        self.client_name = client_name
        self.app_version = app_version
        self._request_address_val = ''

    @property
    def _request_address(self):
        ''' Get address of POST request to the service '''
        if not self._request_address_val:
            template = (
                'https://sb-ssl.google.com/safebrowsing/api/lookup'
                '?client={0}&key={1}&appver={2}&pver={3}'
            )
            self._request_address_val = template.format(
                self.client_name,
                self.api_key,
                self.app_version,
                self.protocol_version
            )
        return self._request_address_val

    def _query_once(self, urls):
        ''' Perform a single POST request using lookup API

        :param urls: a sequence of urls to put in request body
        :returns: a response object
        :raises UnathorizedAPIKeyError: when the API key for this instance
        is not valid
        :raises HTTPError: if the HTTPError was raised for a HTTP code
        other than 401, the exception is reraised
        '''
        request_body = '{}\n{}'.format(len(urls), '\n'.join(urls))
        response = post(self._request_address, request_body)
        try:
            response.raise_for_status()
        except HTTPError as error:
            if response.status_code == 401:
                msg = 'The API key is not authorized'
                raise_from(UnathorizedAPIKeyError(msg), error)
            else:
                raise
        return response

    def _query(self, urls):
        ''' Test urls for being listed by the service

        :param urls: a sequence of urls  to be tested
        :returns: a tuple containing chunk of urls and a response pertaining
          to them if the code of response was 200, which means at least one
          of the queried URLs is matched in either the phishing, malware,
           or unwanted software lists.
        '''
        urls = list(set(urls))
        for i in range(0, len(urls), self.max_urls_per_request):
            chunk = urls[i:i+self.max_urls_per_request]
            response = self._query_once(chunk)
            if response.status_code == 200:
                yield chunk, response

    @accepts_valid_urls
    def any_match(self, urls):
        ''' Check if the service recognizes any of given urls as spam

        :param urls: a sequence of urls to be tested
        :returns: True if any of the urls was recognized as spam
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        return any(self._query(urls))

    def _get_match_and_classification(self, urls):
        ''' Get classification for all matching urls

        :param urls: a sequence of urls to test
        :return: a tuple containing matching url and classification
        string pertaining to it
        '''
        for url_list, response in self._query(urls):
            classification_set = response.text.splitlines()
            for url, _class in zip(url_list, classification_set):
                if _class != 'ok':
                    yield url, _class

    @accepts_valid_urls
    def lookup_matching(self, urls):
        ''' Get items for all listed urls

        :param urls: a sequence of urls to be tested
        :returns: objects representing listed urls
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        for url, _class in self._get_match_and_classification(urls):
            classification = set(_class.split(','))
            yield AddressListItem(url, self, classification)

    @accepts_valid_urls
    def filter_matching(self, urls):
        ''' Get all listed urls

        :param urls: a sequence of urls to be tested
        :returns: spam urls
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        for url, _ in self._get_match_and_classification(urls):
            yield url


SPAMHAUS_XBL_CLASSIFICATION = (
    'CBL (3rd party exploits such as proxies, trojans, etc.)'
)
SPAMHAUS_PBL_CLASSIFICATION = (
    'End-user Non-MTA IP addresses set by ISP outbound mail policy'
)

SPAMHAUS_ZEN_CLASSIFICATION = {
    2: (
        'Direct UBE sources, spam operations & spam services'
    ),
    3: (
        'Direct snowshoe spam sources detected via automation'
    ),
    4: SPAMHAUS_XBL_CLASSIFICATION,
    5: SPAMHAUS_XBL_CLASSIFICATION,
    6: SPAMHAUS_XBL_CLASSIFICATION,
    7: SPAMHAUS_XBL_CLASSIFICATION,
    10: SPAMHAUS_PBL_CLASSIFICATION,
    11: SPAMHAUS_PBL_CLASSIFICATION
}

SPAMHAUS_ZEN = DNSBL(
    'spamhaus_zen',
    'zen.spamhaus.org',
    SPAMHAUS_ZEN_CLASSIFICATION,
    ip_address
)


SPAMHAUS_DBL_CLASSIFICATION = {
    2: 'spam domain',
    4: 'phishing domain',
    5: 'malware domain',
    6: 'botnet C&C domain',
    102: 'abused legit spam',
    103: 'abused spammed redirector domain',
    104: 'abused legit phishing',
    105: 'abused legit malware',
    106: 'abused legit botnet C&C',
}

SPAMHAUS_DBL = DNSBL(
    'spamhaus_dbl',
    'dbl.spamhaus.org',
    SPAMHAUS_DBL_CLASSIFICATION,
    registered_domain
)

SURBL_MULTI_CLASSIFICATION = {
    2: 'deprecated (previously SpamCop web sites)',
    4: 'listed on WS (will migrate to ABUSE on 1 May 2016)',
    8: 'phishing',
    16: 'malware',
    32: 'deprecated (previously AbuseButler web sites)',
    64: 'spam and other abuse sites: (previously jwSpamSpy + Prolocation'
        ' sites, SpamCop web sites, AbuseButler web sites)',
    128: 'Cracked sites'
}

SURBL_MULTI = BitmaskingDNSBL(
    'surbl_multi',
    'multi.surbl.org',
    SURBL_MULTI_CLASSIFICATION,
    registered_domain_or_ip
)



# -*- coding: utf-8 -*-
'''
This module contains unit tests for functions and classes provided by
spam_lists.host_list module
'''
from __future__ import unicode_literals

from spam_lists.host_list import HostList
from test.compat import unittest, Mock, patch
from test.unit.common_definitions import (
    HostListTestMixin, host_list_host_factory
)


class HostListTest(HostListTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for HostList class

    HostList does not provide implementation of some methods it uses.
    These methods are ought to be implemented by its subclasses. Here,
    we mock these methods so that HostList can be tested.

    :var listed_hosts: a list of all host values assumed to be listed for
    a given test
    :var host_factory_mock: a mocked implementation of host factory
    used by tested instance. Uses host_list_host_factory as its implementation
    :var tested_instance: an instance of tested class
    :var _contains_patcher: a patcher for HostList._contains method
    :var _contains_mock: a mock for HostList._contains method.
    :var host_data_getter_patcher: a patcher for
    HostList._get_match_and_classification method
    :var host_data_getter_mock: a mock for
    HostList._get_match_and_classification method. Uses
     host_list_host_factory as its implementation.
    '''
    def setUp(self):
        self.listed_hosts = []
        self.host_factory_mock = Mock()
        self.host_factory_mock.side_effect = host_list_host_factory
        self.tested_instance = HostList(self.host_factory_mock)
        self._contains_patcher = patch(
            'spam_lists.host_list.HostList._contains'
        )
        self._contains_mock = self._contains_patcher.start()
        self._contains_mock.side_effect = lambda h: h in self.listed_hosts
        host_data_getter_name = (
            'spam_lists.host_list.HostList._get_match_and_classification'
        )
        self.host_data_getter_patcher = patch(host_data_getter_name)
        self.host_data_getter_mock = self.host_data_getter_patcher.start()

        def _get_match_and_classification(host):
            if host in self.listed_hosts:
                return host, self.classification
            return None, None
        self.host_data_getter_mock.side_effect = _get_match_and_classification

    def tearDown(self):
        self._contains_patcher.stop()
        self.host_data_getter_patcher.stop()

    def _set_matching_hosts(self, matching_hosts):
        self.listed_hosts = [self.host_factory_mock(mh)
                             for mh in matching_hosts]


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-

'''
This module contains classes of objects representing
custom host collections and their dependencies
'''
from __future__ import unicode_literals
from bisect import bisect_right

from .host_list import HostList
from .structures import hostname_or_ip


class BaseHostCollection(HostList):
    ''' Base class for containers storing ip addresses
    and domain names
    '''
    def __init__(
            self,
            identifier,
            classification,
            hosts=None,
            host_factory=hostname_or_ip):
        ''' Create new instance

        :param identifier: an identifier of this instance of host collection
        :param classification: a list or tuple containing strings representing
        types of items, assigned to each element of the collection
        :param hosts: an object storing ip adresses and hostnames.
        It must have the following methods:
        * __getitem__
        * __len__
        * pop
        * append
        :param host_factory: a callable used to create hosts objects stored
        in the collection or representing values searched in it.
        '''
        self.identifier = identifier
        self.classification = set(classification)
        self.hosts = hosts if hosts is not None else []
        super(BaseHostCollection, self).__init__(host_factory)

    def __len__(self):
        return len(self.hosts)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return self.__class__(
                self.identifier,
                self.classification,
                self.hosts[index],
                self._host_factory
            )
        return self._host_factory(self.hosts[index])

    def _contains(self, host_object):
        match = self._get_match(host_object)
        return match is not None

    def _get_match_and_classification(self, host_object):
        match = self._get_match(host_object)
        _class = None if match is None else self.classification
        return match, _class

    def add(self, host_value):
        ''' Add the given value to collection

        :param host: an ip address or a hostname
        :raises InvalidHostError: raised when the given value
        is not a valid ip address nor a hostname
        '''
        host_obj = self._host_factory(host_value)
        if self._get_match(host_obj) is not None:
            return
        self._add_new(host_obj)

    def _add_new(self, host_object):
        ''' Add a new host to the collection

        A new host is defined as a value not currently listed
        (in case of both hostnames and ip) or not currently
        covered by another value (in case of hostnames, which
        could be covered by their parent domain).

        Before a new hostname can be added, all its subdomains
        already present in the collection must be removed.

        :param host_obj: an object representing value to be added.
        It is assumed that, during execution of this method,
        the value to be added is not currently listed.
        '''
        raise NotImplementedError


class HostCollection(BaseHostCollection):
    ''' Provides a container for ip addresses and domain names.

    May be used as a local whitelist or blacklist.
    '''
    def _get_match(self, host_object):
        for val in self:
            if host_object.is_match(val):
                return val

    def _add_new(self, host_obj):
        for i, listed_obj in enumerate(self):
            if listed_obj.is_subdomain(host_obj):
                self.hosts.pop(i)
        self.hosts.append(host_obj.to_unicode())


class SortedHostCollection(BaseHostCollection):
    ''' A container for ip addresses and hostnames,
    keeping the items in sorted order.
    '''
    def _get_insertion_point(self, host_obj):
        return bisect_right(self, host_obj)

    def _get_match(self, host_object):
        ''' Get an item matching given host object

        The item may be either a parent domain or identical value.
        Parent domains and existing identical values always precede
        insertion point for given value - therefore, we treat
        an item just before insertion point as potential match.

        :param host_object: an object representing ip address
        or hostname whose match we are trying to find
        '''
        i = self._get_insertion_point(host_object)
        potential_match = None
        try:
            potential_match = self[i-1]
        except IndexError:
            pass

        if host_object.is_match(potential_match):
            return potential_match
        return None

    def _add_new(self, host_object):
        ''' Add a new host to the collection

        Before a new hostname can be added, all its subdomains
        already present in the collection must be removed.
        Since the collection is sorted, we can limit our
        search for them to slice of the collection starting
        from insertion point and ending with the last subdomain
        detected
        :param host_obj: an object representing value to be added.
        It is assumed that, during execution of this method,
        the value to be added is not currently listed.
        '''
        i = self._get_insertion_point(host_object)

        for listed in self[i:]:
            if not listed.is_subdomain(host_object):
                break
            self.hosts.pop(i)

        self.hosts.insert(i, host_object.to_unicode())


# -*- coding: utf-8 -*-
'''
====================
spam-lists library
====================

spam-lists provides common interfaces for querying whitelists and
blacklists containing web addresses: hostnames, IP addresses or urls.

It supports the following third party services:

    * `Spamhaus ZEN https://www.spamhaus.org/zen/`_
    * `Spamhaus DBL https://www.spamhaus.org/dbl/`_
    * `SURBL http://www.surbl.org/`_
    * `hpHosts http://www.hosts-file.net/`_
    * `Google Safe Browsing Lookup API https://developers.google.com/
    safe-browsing/lookup_guide`_

In addition, it provides a HostCollection class whose instances can be
used as custom host whitelists or blacklists.

Python versions: 2.7.x, 3.4.x and greater are supported.

************
API:
************

The services and custom web address lists can be queried using
the following interfaces:

    * host list interface, containing methods that receive values
    representing hostnames or IP addresses for which we query
    service represented by an object.

    The methods returns values depending on the given value being
    listed (or not) by the service.

    The interface contains the following methods:

        * __contains__: tests membership of given host in
        the host list. Returns a boolean value.
        * lookup: returns an object representing given item. If it is
        listed, an instance of spam_lists.structures.AddressListItem is
        returned, otherwise the method returns None.

    * url tester interface, containing methods that receive an iterable
    containing urls for which we query service(s) represented by
    an object.

    The methods return values depending on recognizing (or not) any of
    the url values as matching criteria used by the service
    (hostnames, IP addresses, etc.)

    The interface contains the following methods:

        * any_match: checks if any of the urls is recognized as
        a match. Returns a boolean value
        * filter_matching: returns a generator yielding matching urls
        * lookup_matching: returns a generator yielding instances of
        spam_lists.structures.AddressListItem representing matching
        criteria for matching urls.

The following objects are part of the API:

:var SPAMHAUS_DBL: Spamhaus DBL service client, implementing host list
and url tester interfaces.

:var SPAMHAUS_ZEN: Spamhaus ZEN service client, implementing host list
and url tester interfaces.

:var SURBL_MULTI: SURBL service client, implementing host list and url
tester interfaces.

:var HpHosts: a class of objects used as clients for hpHosts service,
implementing host list and url tester interfaces.

:var GoogleSafeBrowsing: a class of objects used as clients for
Google Safe Browsing Lookup API service, implementing url tester
interface.

:var HostCollection: a class of objects representing custom host lists,
implementing host list and url tester interfaces.

:var SortedHostCollection: a class of objects representing custom
sorted host lists, implementing host list and url tester interfaces.

:var UrlTesterChain: a class of objects representing composite
url testers, created by providing objects with url tester methods
as arguments to constructor. It implements url tester interface.

:var GeneralizedUrlTester: a class adding url whitelist and redirect
resolution to a url tester to be used. Instances are created by calling
the constructor with the following arguments:

    * the url tester object
    * (optionally) a whitelist object implementing filter_matching
    method of the url tester interface.
    * (optionally) an instance of redirect url resolver to use

GeneralizedUrlTester implements an interface similar to the url tester
interface, with methods any_match, filter_matching and lookup_matching
all receiving an additional argument: a boolean value specifying if
we should include results of redirect url resolution for given urls in
the data set for which we query service(s) represented by
the url tester.

:copyright: (c) 2016 by Piotr Rusin.
:license: Apache 2.0, see LICENSE for more details.
'''
from __future__ import unicode_literals

from .clients import (
    SPAMHAUS_DBL, SPAMHAUS_ZEN, SURBL_MULTI, HpHosts, GoogleSafeBrowsing
)
from .host_collections import HostCollection, SortedHostCollection
from .composites import UrlTesterChain, GeneralizedUrlTester

__title__ = 'spam-lists'
__version__ = '1.0.0b4'
__author__ = 'Piotr Rusin'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2016 Piotr Rusin'



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from dns.resolver import NXDOMAIN
from future.moves.urllib.parse import urlparse, parse_qs
from nose_parameterized import parameterized
from requests.exceptions import HTTPError

from spam_lists.exceptions import UnathorizedAPIKeyError, UnknownCodeError
from spam_lists.clients import (
    DNSBL, GoogleSafeBrowsing, HpHosts, BitmaskingDNSBL
)
from test.compat import unittest, Mock, patch
from test.unit.common_definitions import (
    HostListTestMixin, host_list_host_factory, UrlTesterTestMixin
)


class DNSQuerySideEffects(object):
    def __init__(self, expected_query_names, last_octet=2):
        self.expected_query_names = expected_query_names
        self.last_octet = last_octet

    def __call__(self, query_name):
        if query_name in self.expected_query_names:
            dns_answer_mock = Mock()
            return_value = '121.0.0.{}'.format(self.last_octet)
            dns_answer_mock.to_text.return_value = return_value
            return [dns_answer_mock]
        raise NXDOMAIN


class DNSBLTestMixin(HostListTestMixin):
    ''' Tests for DNSBL class

    This test case adds additional test method to the ones inherited
    from HostListTestMixin: test_code_error_raised_by, which
    tests methods using return code of a DNSBL service (DNSBL.lookup
    and DNSBL.lookup_matching) for their behaviour for cases of an
    unknown integer code being returned.

    :var query_domain_str: a string used as a suffix for DNS queries
    to a service
    :var host_with_unknown_code: a host value used by the additional
    test method (test_code_error_raised_by)
    :var host_factory_mock: a mocked implementation of host factory
    used by tested instance. Uses host_list_host_factory as its implementation
    :var dnsbl_factory: constructor of instance of tested class
    :var tested_instance: an instance of tested class
    :var dns_query_patcher: an object used for patching query function
     used by DNSBL instance.
    :var dns_query_mock: a mocked implementation of the query function
    '''
    query_domain_str = 'test.query.domain'
    host_with_unknown_code = 'hostwithunknowncode.com'

    def setUp(self):
        self.host_factory_mock = Mock()
        self.host_factory_mock.side_effect = host_list_host_factory
        classification_map = {}
        for i, k in enumerate(self.classification, 1):
            classification_map[2**i] = k
        self.tested_instance = self.dnsbl_factory(
            'test_service',
            self.query_domain_str,
            classification_map,
            self.host_factory_mock
        )
        self.dns_query_patcher = patch('spam_lists.clients.query')
        self.dns_query_mock = self.dns_query_patcher.start()
        self.dns_query_mock.side_effect = DNSQuerySideEffects([])

    def tearDown(self):
        self.dns_query_patcher.stop()

    def _set_matching_hosts(self, hosts):
        host_objects = [self.host_factory_mock(h) for h in hosts]
        query_names = [h.relative_domain.derelativize()
                       for h in host_objects]
        self.dns_query_mock.side_effect.expected_query_names = query_names

    @parameterized.expand([
        ('lookup', host_with_unknown_code),
        (
            'lookup_matching',
            ['http://'+host_with_unknown_code]
        )
    ])
    def test_code_error_raised_by(self, function_name, tested_value):
        self._set_matching_hosts([self.host_with_unknown_code])
        self.dns_query_mock.side_effect.last_octet = 14

        def function(hosts):
            func = getattr(self.tested_instance, function_name)
            return list(func(hosts))

        self.assertRaises(UnknownCodeError, function, tested_value)


class DNSBLTest(DNSBLTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    dnsbl_factory = DNSBL


class BitmaskingDNSBLTest(DNSBLTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    dnsbl_factory = BitmaskingDNSBL


def create_hp_hosts_get(classification, listed_hosts):
    ''' Get a function to replace get function used by HpHosts

    :param classification: a classification for given hosts
    :param listed_hosts: listed hosts for generating responses
    :returns: a function providing side effects of Mock instance
    for the get function
    '''
    class_str = ','.join(classification)

    def hp_hosts_get(url):
        ''' Get mock representing response object for GET request

        :param url: a request address
        :returns: a Mock instance representing response object expected
        by HpHosts
        '''
        query_string = urlparse(url).query
        query_data = parse_qs(query_string)
        content = 'Not Listed'
        host = query_data['s'][0]
        if host in listed_hosts:
            content = 'Listed,{}'.format(class_str)
        response = Mock()
        response.text = content
        return response
    return hp_hosts_get


class HpHostsTest(HostListTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for HpHosts class

    :var listed_hosts: a list of host values assumed to be listed
    for tests
    :var get_patcher: an object used for patching get function used
     by HpHosts instance.
    :var tested_instance: an instance of tested class
    :var get_mock: a mocked implementation of the get function. Uses
    a function returned by create_hp_hosts_get for given classification
    and list of hosts
    :var host_factory_mock: a mocked implementation of
     host factory used by tested instance. Uses host_list_host_factory
      as its implementation
    '''
    @classmethod
    def setUpClass(cls):
        cls.tested_instance = HpHosts('spam_lists_test_suite')

    def setUp(self):
        self.listed_hosts = []
        self.get_patcher = patch('spam_lists.clients.get')
        self.get_mock = self.get_patcher.start()
        self.get_mock.side_effect = create_hp_hosts_get(
            self.classification,
            []
        )
        self.host_factory_mock = Mock()
        self.tested_instance = HpHosts('spam_lists_test_suite')
        self.tested_instance._host_factory = self.host_factory_mock
        self.host_factory_mock.side_effect = host_list_host_factory

    def tearDown(self):
        self.get_patcher.stop()

    def _set_matching_hosts(self, hosts):
        side_effect = create_hp_hosts_get(
            self.classification,
            hosts
        )
        self.get_mock.side_effect = side_effect


def create_gsb_post(expected_401, spam_urls, classification):
    ''' Get mock for post function used by GoogleSafeBrowsing

    :param expected_401: if True, the code of response mock returned
    by the returned function will be 401
    :param spam_urls: a list of urls to be recognized as spam
    :param classification: a classification used for spam urls
    :returns: mocked implementation of post function
    '''
    def post(_, body):
        ''' Get mock of a response to a POST query to GSB lookup API

        :param body: a request body
        :returns: a Mock instance representing the response. Properties
        of the object depend on external values provided by the creator
        of the method: expected_401, spam_urls and classification
        '''
        response = Mock()
        if expected_401:
            response.status_code = 401
            response.raise_for_status.side_effect = HTTPError
        else:
            urls = body.splitlines()[1:]
            classes = ['ok' if u not in spam_urls else
                       ','.join(classification) for u in urls]
            response.text = '\n'.join(classes)
            code = 200 if spam_urls else 204
            response.status_code = code
        return response
    return post


class GoogleSafeBrowsingTest(UrlTesterTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for GoogleSafeBrowsing class

    This class adds an additional test method to the ones provided
    by UrlTesterTestMixin: test_unathorized_query_with. This method
    is used to test methods of GoogleSafeBrowsing class for expected
    behaviour while calling Google Safe Browsing lookup API with
    an unathorized API key

    :var tested_instance: an instance of tested class
    :var post_patcher: an object used for patching post function used
    by GoogleSafeBrowsing instance
    :var mocked_post: a mocked implementation of the post function
    for the tested instance. Uses a function returned by
     create_gsb_post function as its implementation.
    '''
    def _get_expected_items_for_urls(self, urls):
        return self._get_expected_items(urls)

    @classmethod
    def setUpClass(cls):
        cls.tested_instance = GoogleSafeBrowsing(
            'test_client',
            '0.1',
            'test_key'
        )

    def _set_up_post_mock(self, spam_urls, error_401_expected=False):
        side_efect = create_gsb_post(
            error_401_expected,
            spam_urls,
            self.classification
        )
        self.mocked_post.side_effect = side_efect

    def setUp(self):
        self.post_patcher = patch('spam_lists.clients.post')
        self.mocked_post = self.post_patcher.start()

    def tearDown(self):
        self.post_patcher.stop()

    def _set_matching_urls(self, urls):
        self._set_up_post_mock(urls)

    @parameterized.expand([
        ('any_match'),
        ('lookup_matching'),
        ('filter_matching')
    ])
    def test_unathorized_query_with(self, function_name):
        tested_function = getattr(self.tested_instance, function_name)

        def called_function(urls):
            return list(tested_function(urls))
        self._set_up_post_mock([], error_401_expected=True)
        self.assertRaises(
            UnathorizedAPIKeyError,
            called_function,
            self.valid_urls
        )


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
'''
This module contains all classes of exceptions raised
by the library
'''
from __future__ import unicode_literals


class SpamListsError(Exception):
    '''There was an error during testing a url or host'''


class SpamListsValueError(SpamListsError, ValueError):
    '''An inapropriate value was used in spam-lists library '''


class UnknownCodeError(SpamListsError, KeyError):
    '''The classification code from the service was not recognized'''


class UnathorizedAPIKeyError(SpamListsValueError):
    '''The API key used to query the service was not authorized'''


class InvalidHostError(SpamListsValueError):
    '''The value is not a valid host'''


class InvalidIPError(InvalidHostError):
    ''' The value is not a valid IP address'''


class InvalidIPv4Error(InvalidIPError):
    '''The value is not a valid IPv4 address'''


class InvalidIPv6Error(InvalidIPError):
    '''The value is not a valid IPv6 address'''


class InvalidHostnameError(InvalidHostError):
    '''The value is not a valid hostname'''


class InvalidURLError(SpamListsValueError):
    '''The value is not a valid url'''



# -*- coding: utf-8 -*-
'''
This module contains integration tests for supported
third party services

The purpose of the tests is to signal changes in the services that
require changes in the implementation of their clients.
'''
from __future__ import unicode_literals

import os.path

from builtins import object  # pylint: disable=redefined-builtin
import tldextract
from validators import ipv6

from spam_lists.clients import (
    SPAMHAUS_ZEN, SPAMHAUS_ZEN_CLASSIFICATION, SPAMHAUS_DBL,
    SPAMHAUS_DBL_CLASSIFICATION, SURBL_MULTI, SURBL_MULTI_CLASSIFICATION
)
from spam_lists.clients import HpHosts, GoogleSafeBrowsing
from spam_lists.structures import AddressListItem
from test.compat import unittest


def ip_or_registered_domain(host):
    registered_domain = tldextract.extract(host).registered_domain
    return host if not registered_domain else registered_domain


def url_from_host(host):
    if ipv6(host):
        host = '['+host+']'
    return 'http://'+host


def get_classification(classification, return_codes):
    return set(v for k, v in list(classification.items()) if k in return_codes)


class UrlTesterClientTestMixin(object):
    '''  A class containing integration test methods for
    url tester clients

    :var tested_client: an instance of client to be tested
    :var urls_without_listed: urls without values listed by the service
    to be queried
    :var urls_with_listed: urls with values listed by the service
    to be queried
    :var listed_url: a url listed (or: with a host listed) by the service
    to be queried
    :var listed_item: an instance of AddressListItem representing
    an item listed by the service to be queried
    '''
    def test_any_match_for_not_listed(self):
        actual = self.tested_client.any_match(self.urls_without_listed)
        self.assertFalse(actual)

    def test_any_match_for_listed(self):
        actual = self.tested_client.any_match(self.urls_with_listed)
        self.assertTrue(actual)

    def test_filter_matching_not_listed(self):
        generator = self.tested_client.filter_matching(
            self.urls_without_listed
        )
        actual = list(generator)
        self.assertCountEqual([], actual)

    def test_filter_matching_for_listed(self):
        expected = [self.listed_url]
        filter_matching = self.tested_client.filter_matching
        actual = list(filter_matching(self.urls_with_listed))
        self.assertCountEqual(expected, actual)

    def test_lookup_matching_not_listed(self):
        generator = self.tested_client.lookup_matching(
            self.urls_without_listed
        )
        actual = list(generator)
        self.assertCountEqual([], actual)

    def test_lookup_matching_for_listed(self):
        expected = [self.listed_item]
        lookup_matching = self.tested_client.lookup_matching
        actual = list(lookup_matching(self.urls_with_listed))
        self.assertCountEqual(expected, actual)


class HostListClientTestMixin(UrlTesterClientTestMixin):
    '''  A class containing integration test methods for
    host list clients

    :var listed: an item listed by a service to be queried
    :var not_listed: an item not listed by a service to be queried
    :var tested_client: an instance of client to be tested
    :var urls_without_listed: urls without values listed by the service
    to be queried
    :var urls_with_listed: urls with values listed by the service
    to be queried
    :var listed_url: a url listed (or: with a host listed) by the service
    to be queried
    :var listed_item: an instance of AddressListItem representing
    an item listed by the service to be queried
    '''
    @classmethod
    def setUpClass(cls):
        cls.listed_url = url_from_host(cls.listed)
        cls.not_listed_url = url_from_host(cls.not_listed)
        cls.urls_with_listed = cls.not_listed_url, cls.listed_url
        cls.urls_without_listed = (
            cls.not_listed_url,
            url_from_host(cls.not_listed_2)
        )
        cls.listed_item = AddressListItem(
            ip_or_registered_domain(cls.listed),
            cls.tested_client,
            cls.classification
        )

    def test__contains__for_not_listed(self):
        actual = self.not_listed in self.tested_client
        self.assertFalse(actual)

    def test_contains_for_listed(self):
        actual = self.listed in self.tested_client
        self.assertTrue(actual)

    def test_lookup_for_not_listed(self):
        actual = self.tested_client.lookup(self.not_listed)
        self.assertIsNone(actual)

    def test_lookup_for_listed(self):
        actual = self.tested_client.lookup(self.listed)
        self.assertEqual(self.listed_item, actual)


REASON_TO_SKIP = (
    'These tests are expected to fail frequently for users of public'
    ' DNS resolvers:'
    ' https://www.spamhaus.org/faq/section/DNSBL%20Usage#261'
)


# @unittest.skip(REASON_TO_SKIP)
class SpamhausZenTest(HostListClientTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    tested_client = SPAMHAUS_ZEN
    listed = '127.0.0.2'
    not_listed = '127.0.0.1'
    not_listed_2 = '8.8.8.8'
    classification = get_classification(
        SPAMHAUS_ZEN_CLASSIFICATION,
        [2, 4, 10]
    )


# @unittest.skip(REASON_TO_SKIP)
class SpamhausDBLTest(HostListClientTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    tested_client = SPAMHAUS_DBL
    listed = 'dbltest.com'
    not_listed = 'example.com'
    not_listed_2 = 'google.com'
    classification = get_classification(
        SPAMHAUS_DBL_CLASSIFICATION,
        [2]
    )


class SURBLTest(HostListClientTestMixin):
    tested_client = SURBL_MULTI
    classification = get_classification(
        SURBL_MULTI_CLASSIFICATION,
        [2, 4, 8, 16, 32, 64]
    )


class SURBLMultiIPTest(SURBLTest, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    listed = '127.0.0.2'
    not_listed = '127.0.0.1'
    not_listed_2 = '8.8.8.8'


class SURBLMultiDomainTest(SURBLTest, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    listed = 'surbl-org-permanent-test-point.com'
    not_listed = 'test.com'
    not_listed_2 = 'google.com'

HP_HOSTS = HpHosts('spam-lists-test-suite')


class HpHostsIPTest(HostListClientTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    listed = '174.36.207.146'
    not_listed = '64.233.160.0'
    not_listed_2 = '2001:ddd:ccc:123::55'
    tested_client = HP_HOSTS
    classification = set()


class HpHostsDomainTest(HostListClientTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    listed = 'ecardmountain.com'
    not_listed = 'google.com'
    not_listed_2 = 'microsoft.com'
    tested_client = HP_HOSTS
    classification = set(['EMD'])


GSB_API_KEY_FILE = os.path.join(
    os.path.dirname(__file__),
    'google_safe_browsing_api_key.txt'
)
try:
    with open(GSB_API_KEY_FILE, 'r') as key_file:
        SAFE_BROWSING_API_KEY = key_file.readline().rstrip()
except IOError:
    SAFE_BROWSING_API_KEY = None


REASON_TO_SKIP_GSB_TEST = (
    'No api key provided. Provide the key in file: {}'.format(GSB_API_KEY_FILE)
    )


@unittest.skipIf(not SAFE_BROWSING_API_KEY, REASON_TO_SKIP_GSB_TEST)
class GoogleSafeBrowsingTest(UrlTesterClientTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    listed_url = 'http://www.gumblar.cn/'
    not_listed_url = 'http://www.google.com/'
    not_listed_url_2 = 'https://github.com/'
    urls_with_listed = not_listed_url, listed_url
    urls_without_listed = not_listed_url, not_listed_url_2

    @classmethod
    def setUpClass(cls):
        cls.tested_client = GoogleSafeBrowsing(
            'spam-lists-test-suite',
            '0.5',
            SAFE_BROWSING_API_KEY
        )
        cls.listed_item = AddressListItem(
            cls.listed_url,
            cls.tested_client,
            set(['malware'])
        )


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
'''
A module providing a compatibility layer for python 2 and 3.

Most of the compatibility-related features are already provided
by future and other modules. This module contains compatibility
features to be shared by spam-lists and test packages of
this project, like definitions depending on the version
of the interpreter, complex imports that may result in
errors, etc.

For now, it contains only code responsible for importing
functools.lru_cache or cachetools.func.lru_cache
'''
# pylint: disable=unused-import
# pylint: disable=import-error

try:
    from functools import lru_cache  # @NoMove
except ImportError:
    from cachetools.func import lru_cache  # @NoMove @UnusedImport



# -*- coding: utf-8 -*-

'''
This module contains various utilities to be used to create
composite spam url checkers.
'''

from __future__ import unicode_literals

from builtins import object  # pylint: disable=redefined-builtin
from requests import Session
from requests.exceptions import (
    ConnectionError, InvalidSchema, InvalidURL, Timeout
)

from .exceptions import InvalidURLError
from .validation import is_valid_url


class CachedIterable(object):
    '''An iterable returning items from an iterator
    and caching them for future runs

    Items are returned in fixed order.
    '''
    def __init__(self, iterator, initial_cache=None):
        ''' Constructor

        :param iterator: an iterator, wrom which items
         will be returned and cached
        :param initial_cache: values added to cache before
         the first run
        '''
        if initial_cache is None:
            initial_cache = []
        self._cache = list(initial_cache)
        self._iterator = iterator

    def __iter__(self):
        for i in self._cache:
            yield i
        for i in self._iterator:
            self._cache.append(i)
            yield i


class RedirectUrlResolver(object):
    ''' A class used for getting all redirect urls for
    given url

    The urls include:
    * url addresses of all responses acquired for a HEAD request in its
     response history
    * value of location header for the last response, if it is
    a valid url but we still couldn't get a response for it
    '''
    def __init__(self, requests_session=Session()):
        '''
        Constructor

        :param requests_session: a session object implementing
        methods:
        * head(url) (for HEAD request)
        * resolve_redirects(response, request)
        '''
        self.session = requests_session

    def get_locations(self, url):
        ''' Get valid location header values from
        responses for given url

        :param url: a url address. If a HEAD request sent to it
        fails because the address has invalid schema, times out
        or there is a connection error, the generator yields nothing
        :returns: valid redirection addresses. If a request for
        a redirection address fails, and the address is still a valid
        url string, it's included as the last yielded value. If it's
        not, the previous value is the last one.
        :raises ValuError: if the argument is not a valid url
        '''
        if not is_valid_url(url):
            raise InvalidURLError('{} is not a valid url'.format(url))
        try:
            response = self.session.head(url)
        except (ConnectionError, InvalidSchema, Timeout):
            raise StopIteration
        try:
            generator = self.session.resolve_redirects(
                response,
                response.request
            )
            for response in generator:
                yield response.url
        except InvalidURL:
            pass
        except (ConnectionError, InvalidSchema, Timeout) as error:
            last_url = response.headers['location']
            if isinstance(error, Timeout) or is_valid_url(last_url):
                yield last_url

    def get_new_locations(self, urls):
        ''' Get valid location header values for all given urls

        The returned values are new, that is: they do not repeat any
        value contained in the original input. Only unique values
        are yielded.

        :param urls: a list of url addresses
        :returns: valid location header values from responses
        to the urls
        '''
        seen = set(urls)
        for i in urls:
            for k in self.get_locations(i):
                if k not in seen:
                    seen.add(k)
                    yield k

    def get_urls_and_locations(self, urls):
        ''' Get urls and their redirection addresses

        :param urls: a list of url addresses
        :returns: an instance of CachedIterable containing given urls
        and valid location header values of their responses
        '''
        location_generator = self.get_new_locations(urls)
        initial_cache = list(set(urls))
        return CachedIterable(location_generator, initial_cache)


class UrlTesterChain(object):
    '''A url tester using a sequence of other url testers'''
    def __init__(self, *url_testers):
        '''Constructor

        :param url_testers: a list of objects having any_match(urls)
        and lookup_matching(urls) methods
        '''
        self.url_testers = list(url_testers)

    def any_match(self, urls):
        ''' Check if any of given urls is a match

        :param urls: a sequence of urls to be tested
        :returns: True if any of the urls is a match
        '''
        return any(t.any_match(urls) for t in self.url_testers)

    def lookup_matching(self, urls):
        '''Get objects representing match criteria
        (hosts, whole urls, etc) for given urls

        :param urls: an iterable containing urls
        :returns: items representing match criteria
        '''
        for tester in self.url_testers:
            for item in tester.lookup_matching(urls):
                yield item

    def filter_matching(self, urls):
        ''' Get those of given ruls that match listing criteria
        (hosts, whole urls, etc.)

        :param urls: an iterable containing urls
        :returns: matching urls
        '''
        seen = set()
        urls = set(urls)
        for tester in self.url_testers:
            urls = urls - seen
            for url in tester.filter_matching(urls):
                if url not in seen:
                    seen.add(url)
                    yield url


class GeneralizedUrlTester(object):
    ''' A url tester using redirect resolution, whitelist
    and another url tester
    '''
    def __init__(self, url_tester, whitelist=None,
                 redirect_resolver=RedirectUrlResolver()):
        ''' Constructor

        :param url_tester: an object with any_match, filter_matching
        and lookup_matching methods that can be used for testing urls
        :param whitelist: an object with a filter_matching method, used
        for filtering urls to be tested against the url_tester
        :param redirect_resolver: an object used for getting valid
        location header values to test them with the other url values.
        '''
        self.url_tester = url_tester
        self.whitelist = whitelist
        self.redirect_resolver = redirect_resolver

    def _get_results_for(self, function, urls, resolve_redirects):
        ''' Get results of given function for given arguments

        :param function: a function to be called
        :param urls: an iterable containing initial url values
        :param resolve_redirects: a boolean value. If True, all valid
        redirect location values will be resolved for given urls and
        tested with them
        '''
        urls_to_test = urls
        if resolve_redirects:
            urls_to_test = self.redirect_resolver.get_urls_and_locations(urls)
        if self.whitelist is not None:
            generator = self.whitelist.filter_matching(urls_to_test)
            urls_to_test = list(set(urls_to_test) - set(generator))
        return function(urls_to_test)

    def any_match(self, urls, resolve_redirects=True):
        ''' Check if any of given urls is a match

        :param urls: an iterable containing initial url values
        :param resolve_redirects: a boolean value. If True, all valid
        redirect location values will be resolved for given urls and
        tested with them
        :returns: True if any of the urls is a match
        '''
        return self._get_results_for(
            self.url_tester.any_match,
            urls,
            resolve_redirects
        )

    def filter_matching(self, urls, resolve_redirects=True):
        ''' Get those of given ruls that match listing criteria
        (hosts, whole urls, etc.)

        :param urls: an iterable containing initial url values
        :param resolve_redirects: a boolean value. If True, all valid
        redirect location values will be resolved for given urls and
        tested with them
        :returns: matching urls
        '''
        return self._get_results_for(
            self.url_tester.filter_matching,
            urls,
            resolve_redirects
        )

    def lookup_matching(self, urls, resolve_redirects=True):
        '''Get objects representing match criteria
        (hosts, whole urls, etc) for given urls

        :param urls: an iterable containing initial url values
        :param resolve_redirects: a boolean value. If True, all valid
        redirect location values will be resolved for given urls and
        tested with them
        :returns: items representing match criteria
        '''
        return self._get_results_for(
            self.url_tester.lookup_matching,
            urls,
            resolve_redirects
        )



# -*- coding: utf-8 -*-

'''
This module contains a definition of HostList - a common base class
for classes representing host lists, like clients of online host
blacklists or custom host whitelists and blacklists.
'''
from __future__ import unicode_literals

# pylint: disable=redefined-builtin
from builtins import object
from future.moves.urllib.parse import urlparse

from .exceptions import InvalidHostError
from .structures import AddressListItem
from .validation import accepts_valid_urls, accepts_valid_host


class HostList(object):
    ''' A class of clients for local or remote host list services '''
    def __init__(self, host_factory):
        ''' Constructor

        :param host_factory: a function responsible for
        creating valid host objects. It may raise InvalidHostError
        (or its subclasses) if a value passed to it is not
        a valid host of type accepted by the factory.
        '''
        self._host_factory = host_factory

    def _contains(self, host_value):
        ''' Check if the service lists an item
        matching given host value

        :param host_value: a host value
        :returns: True if the service lists a matching
        value
        '''
        raise NotImplementedError

    def _get_match_and_classification(self, host_value):
        ''' Get a listed value that matches
        given host value and its classification

        :param host_value: a host value
        :returns: a tuple containing listed item and its classification as
        a tuple containing all classification groups to which the item belongs
        '''
        raise NotImplementedError

    @accepts_valid_host
    def __contains__(self, host_value):
        ''' Check if given host value is listed by the service

        :param host_value: a string representing a valid host
        :returns: True if the host is listed
        :raises InvalidHostError: if the argument is not a valid host string
        '''
        try:
            host_object = self._host_factory(host_value)
        except InvalidHostError:
            return False
        return self._contains(host_object)

    @accepts_valid_host
    def lookup(self, host_value):
        ''' Get an object representing a host value
        matched by this host

        :param host_value: a value of the host of a type
        that can be listed by the service
        :returns: an instance of AddressListItem representing
        a matched value
        :raises InvalidHostError: if the argument is not a valid host string
        '''
        try:
            host_object = self._host_factory(host_value)
        except InvalidHostError:
            return None
        result = self._get_match_and_classification(
            host_object
        )
        host_item, classification = result
        if host_item is not None:
            return AddressListItem(
                host_item.to_unicode(),
                self,
                classification
            )
        return None

    @accepts_valid_urls
    def any_match(self, urls):
        ''' Check if any of given urls has a listed host

        :param urls: an iterable containing urls
        :returns: True if any host is listed
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        return any(urlparse(u).hostname in self for u in urls)

    @accepts_valid_urls
    def lookup_matching(self, urls):
        '''Get objects representing hosts in given urls
        that match listed hosts

        :param urls: an iterable containing urls
        :returns: items representing hosts matching the listed ones
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        hosts = (urlparse(u).hostname for u in urls)
        for val in hosts:
            item = self.lookup(val)
            if item is not None:
                yield item

    @accepts_valid_urls
    def filter_matching(self, urls):
        ''' Get urls with hosts matching listed ones

        :param urls: an iterable containing url addresses to filter
        :returns: a list containing matching urls
        :raises InvalidURLError: if there are any invalid urls in the sequence
        '''
        def is_match(url):
            return urlparse(url).hostname in self
        return (u for u in urls if is_match(u))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from dns import name
from ipaddress import ip_address
from nose_parameterized import parameterized

from spam_lists.exceptions import InvalidHostError
from spam_lists.host_collections import HostCollection, SortedHostCollection
from test.compat import unittest, Mock
from test.unit.common_definitions import (
    TestFunctionDoesNotHandleMixin, host_list_host_factory, HostListTestMixin
)


def get_sorting_key(value):
    ''' Returns a sorting key used for sorting
    host values during test

    :param value: a host value for which we generate key
    '''
    try:
        return ip_address(value)
    except ValueError:
        return name.from_text(value)


def has_to_unicode(value):
    return hasattr(value, 'to_unicode')


def host_collection_host_factory(host):
    host_object = host_list_host_factory(host)
    _str = host_object.to_unicode()

    def test(other):
        return (has_to_unicode(other) and
                other.to_unicode() in _str)
    host_object.is_match.side_effect = test
    host_object.is_subdomain.side_effect = test

    def less_than(other):
        ''' An implementation of __lt__ expected
        from host objects by bisect_right

        :param other: a value to be compared
        '''
        host_object_key = get_sorting_key(_str)
        other_value = other.to_unicode() if has_to_unicode(other) else other
        other_key = get_sorting_key(other_value)
        try:
            result = host_object_key < other_key
        except TypeError:
            result = _str < other_value
        return result

    host_object.__lt__.side_effect = less_than
    return host_object


class HostCollectionBaseTest(
        HostListTestMixin,
        TestFunctionDoesNotHandleMixin,
):
    # pylint: disable=too-many-public-methods
    ''' Tests for subclasses or BaseHostCollection

    :var host_factory_mock: a mocked implementation of
     host factory used by tested instance. Uses
      host_collection_host_factory as its implementation
    :var tested_instance: an instance of tested class
    '''
    valid_urls = ['http://test.com', 'http://127.33.22.11']

    def setUp(self):
        self.host_factory_mock = Mock()
        self.host_factory_mock.side_effect = host_collection_host_factory
        self.tested_instance = self.constructor(
            'test_host_collection',
            self.classification,
            host_factory=self.host_factory_mock
        )

    def test_add_invalid_host(self):
        function = self.tested_instance.add
        self._test_function_does_not_handle(
            InvalidHostError,
            self.host_factory_mock,
            function,
            'invalidhost.com'
        )

    @parameterized.expand(HostListTestMixin.valid_host_input)
    def test_add_for_valid(self, _, value):
        self.tested_instance.add(value)
        self.assertTrue(value in self.tested_instance.hosts)

    def test_add_for_subdomain(self):
        ''' A subdomain to a domain already listed in the collection
        is expected to be ignored when added to the collection '''
        initial_hosts = ['domain.com']
        self._set_matching_hosts(initial_hosts)
        self.tested_instance.add('subdomain.domain.com')
        self.assertCountEqual(initial_hosts, self.tested_instance.hosts)

    def test_add_for_the_same_value(self):
        '''A value being added to the collection is being ignored if it
        already exists in the collection '''
        value = 'domain.com'
        initial_hosts = ['host.com', value]
        self._set_matching_hosts(initial_hosts)
        self.tested_instance.add(value)
        self.assertCountEqual(initial_hosts, self.tested_instance.hosts)

    def test_add_a_superdomain(self):
        ''' A superdomain of a domain listed in the collection
        is expected to replace its subdomain when added '''
        superdomain = 'domain.com'
        subdomain = 'sub.domain.com'
        initial_hosts = ['host1.com', subdomain]
        self._set_matching_hosts(initial_hosts)
        self.tested_instance.add(superdomain)
        initial_hosts.remove(subdomain)
        initial_hosts.append(superdomain)
        self.assertCountEqual(initial_hosts, self.tested_instance.hosts)

    def _set_matching_hosts(self, hosts):
        self.tested_instance.hosts = list(hosts)


class HostCollectionTest(HostCollectionBaseTest, unittest.TestCase):
    constructor = HostCollection


class SortedHostCollectionTest(HostCollectionBaseTest, unittest.TestCase):
    constructor = SortedHostCollection

    def _set_matching_hosts(self, hosts):
        self.tested_instance.hosts = list(hosts)
        self.tested_instance.hosts.sort(key=self.host_factory_mock)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-

'''
This module contains functions responsible for
validating arguments for other functions and
methods provided by the library
'''
from __future__ import unicode_literals

import functools
import re

from future.moves.urllib.parse import urlparse
import validators

from .exceptions import InvalidURLError, InvalidHostError


def is_valid_host(value):
    ''' Check if given value is valid host string

    :param value: a value to test
    :returns: True if the value is valid host string
    '''
    host_validators = validators.ipv4, validators.ipv6, validators.domain
    return any(f(value) for f in host_validators)

URL_REGEX = re.compile(r'^[a-z0-9\.\-\+]*://'  # scheme
                       r'(?:\S+(?::\S*)?@)?'  # authentication
                       r'(?:[^/:]+|\[[0-9a-f:\.]+\])'  # host
                       r'(?::\d{2,5})?'  # port
                       r'(?:[/?#][^\s]*)?'  # path, query or fragment
                       r'$', re.IGNORECASE)


def is_valid_url(value):
    ''' Check if given value is valid url string

    :param value: a value to test
    :returns: True if the value is valid url string
    '''
    match = URL_REGEX.match(value)
    host_str = urlparse(value).hostname
    return match and is_valid_host(host_str)


def accepts_valid_host(func):
    ''' Return a wrapper that runs given method only for arguments
    that are valid host values

    :param func: a method to be wrapped
    :returns: a wrapper that adds argument validation
    '''
    @functools.wraps(func)
    def wrapper(obj, value, *args, **kwargs):
        ''' Run the function and return its return value
        if the value is host - otherwise raise InvalidHostError
        :param obj": an object in whose class f is defined
        :param value: a value expected to be a valid host string
        :returns: a return value of the function f
        :raises InvalidHostError: if the value is not a valid host string
        '''
        if not is_valid_host(value):
            raise InvalidHostError
        return func(obj, value, *args, **kwargs)
    return wrapper


def accepts_valid_urls(func):
    ''' Return a wrapper that runs given method only for arguments
    that are valid url values

    :param func: a method to be wrapped
    :returns: a wrapper that adds argument validation
    '''
    @functools.wraps(func)
    def wrapper(obj, urls, *args, **kwargs):
        '''Run the function and return its return value
         if all given urls are valid - otherwise raise InvalidURLError
        :param obj: an object in whose class f is defined
        :param urls: an iterable containing urls
        :returns: a return value of the function f
        :raises InvalidURLError: if the iterable contains invalid urls
        '''
        invalid_urls = [u for u in urls if not is_valid_url(u)]
        if invalid_urls:
            msg_tpl = 'The values: {} are not valid urls'
            msg = msg_tpl.format(','.join(invalid_urls))
            raise InvalidURLError(msg)
        return func(obj, urls, *args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
'''
This module contains unit tests for functions and classes provided by
spam_lists.validation module
'''
from __future__ import unicode_literals

import unittest

from builtins import object  # pylint: disable=redefined-builtin
from nose_parameterized import parameterized

from spam_lists.exceptions import InvalidURLError, InvalidHostError
from spam_lists.validation import (
    accepts_valid_urls, is_valid_url, accepts_valid_host
)
from test.compat import Mock, patch


class ValidationDecoratorTestMixin(object):
    ''' Provides tests for decorators and wrappers responsible for
     testing validity of arguments of decorated methods

    :var validity_tester_patcher: an object used for patching
     a function responsible for testing validity of arguments of
     a decorated function
    :var validity_tester_mock: a mocked implementation for
     the validity tester
    :var obj: a mock representing object having the method decorated
     by the decorator
    :var function: a mock representing function to be decorated
    :var decorated_function: a result of applying decorator to
     the function

     Additionally, the test cases using this mixin are expected to have
     the following attributes:

    :var exception_type:
    :var decorator:
    :var validity_tester: a fully qualified name of a function used by
     the tested wrapper as argument validator
    '''
    def setUp(self):
        self.validity_tester_patcher = patch(self.validity_tester)
        self.validity_tester_mock = self.validity_tester_patcher.start()
        function = Mock()
        function.__name__ = str('function')
        self.obj = Mock()
        self.function = function
        self.decorated_function = self.decorator(self.function)

    def tearDown(self):
        self.validity_tester_patcher.stop()

    def _test_wrapper_for_valid(self, value):
        self.decorated_function(self.obj, value)
        self.function.assert_called_once_with(self.obj, value)

    def _test_wrapper_for_invalid(self, value):
        self.validity_tester_mock.return_value = False
        self.assertRaises(
            self.exception_type,
            self.decorated_function,
            self.obj,
            value
        )
        self.function.assert_not_called()


class AcceptValidUrlsTest(ValidationDecoratorTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for accepts_valid_urls decorator '''
    exception_type = InvalidURLError
    decorator = staticmethod(accepts_valid_urls)
    validity_tester = 'spam_lists.validation.is_valid_url'

    @parameterized.expand([
        ('hostname', ['https://valid.com']),
        ('ipv4_host', ['http://122.34.59.109']),
        ('ipv6_host', ['http://[2001:db8:abc:123::42]'])
    ])
    def test_for_urls_with_valid(self, _, urls):
        self._test_wrapper_for_valid(urls)

    @parameterized.expand([
        ('invalid_hostname', ['http://-abc.com']),
        ('invalid_schema', ['abc://hostname.com']),
        ('no_schema', ['hostname.com']),
        ('invalid_ipv4', ['http://999.999.999.999']),
        ('invalid_ipv4', ['http://127.0.0.0.1']),
        ('invalid_ipv6', ['http://[2001:db8:abcef:123::42]']),
        ('invalid_ipv6', ['http://[2001:db8:abch:123::42]'])])
    def test_for_urls_with(self, _, urls):
        self._test_wrapper_for_invalid(urls)


class AcceptsValidHostTest(ValidationDecoratorTestMixin, unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for accepts_valid_host decorator '''
    exception_type = InvalidHostError
    decorator = staticmethod(accepts_valid_host)
    validity_tester = 'spam_lists.validation.is_valid_host'

    @parameterized.expand([
        ('hostname', 'valid.com'),
        ('ipv4', '122.34.59.109'),
        ('ipv6', '2001:db8:abc:123::42')
    ])
    def test_for_valid(self, _, value):
        self._test_wrapper_for_valid(value)

    @parameterized.expand([
        ('hostname', '-abc.com'),
        ('ipv4', '999.999.999.999'),
        ('ipv4', '127.0.0.0.1'),
        ('ipv6', '2001:db8:abcef:123::42'),
        ('ipv6', '2001:db8:abch:123::42')
    ])
    def test_for_invalid(self, _, value):
        self._test_wrapper_for_invalid(value)


class IsValidUrlTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for is_valid_url function '''
    @parameterized.expand([
        ('http_scheme', 'http://test.url.com'),
        ('https_scheme', 'https://google.com'),
        ('ftp_scheme', 'ftp://ftp.test.com'),
        ('numeric_hostname', 'http://999.com'),
        ('final_slash', 'https://google.com/'),
        ('path_query_and_fragment', (
            'https://test.domain.com/path/element'
            '?var=1&var_2=3#fragment'
        )),
        ('query', 'http://test.domain.com?var_1=1&var_2=2'),
        ('path', 'http://test.domain.com/path'),
        ('path_and_fragment', 'http://test.domain.com/path#fragment'),
        (
            'query_and_fragment',
            'http://test.domain.com?var_1=1&var_2=2#fragment'
        ),
        ('port', 'https://test.domain.com:123'),
        ('authentication', 'https://abc:def@test.domain.com'),
        ('ipv4', 'http://255.0.0.255'),
        ('ipv6', 'http://[2001:db8:abc:125::45]'),
        ('no_schema', 'test.url.com', False),
        ('invalid_ipv4', 'http://266.0.0.266', False),
        ('invalid_ipv6', 'http://127.0.0.1.1', False),
        ('invalid_port', 'http://test.domain.com:aaa', False),
        ('no_top_level_domain', 'https://testdomaincom', False),
        ('invalid_hostname', 'http://-invalid.domain.com', False)
    ])
    def test_for_url_with(self, _, url, expected=True):
        actual = is_valid_url(url)
        if expected:
            self.assertTrue(actual)
        else:
            self.assertFalse(actual)

if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
'''
This module contains unit tests for functions and classes provided by
spam_lists.composites module
'''
from __future__ import unicode_literals

from collections import defaultdict
from random import shuffle

from builtins import next, range  # pylint: disable=redefined-builtin
from nose_parameterized import parameterized
from requests.exceptions import (
    ConnectionError, InvalidSchema, InvalidURL, Timeout
)

from spam_lists.exceptions import InvalidURLError, UnknownCodeError
from spam_lists.structures import AddressListItem
from spam_lists.composites import (
    RedirectUrlResolver, UrlTesterChain, CachedIterable, GeneralizedUrlTester
)
from test.compat import unittest, Mock, patch, lru_cache, MagicMock
from test.unit.common_definitions import (
    UrlTesterTestBaseMixin, TestFunctionDoesNotHandleMixin
)


def get_response_mock(url):
    ''' Get mock representing response to a request

    :param url: response url
    :returns: an instance of mock representing a response
    '''
    response = Mock()
    response.url = url
    return response


class HeadSideEffects(dict):
    def __call__(self, url):
        return self.get(url)


class ResolveRedirectsSideEffects(object):
    '''' Provides side effects for redirect resolution

    The side effects include both response object mocks and exceptions.

    :var redirect_responses: a dictionary mapping response mocks
    to objects representing response arguments of
    the requests.Session.resolve_redirects method.
    :var exceptions: a dictionary mapping exception types to
    objects representing response arguments of the resolve_redirects
    method
    '''
    def __init__(self):
        self.responses = {}
        self.exceptions = {}

    def __call__(self, response, request):
        yielded = self.responses.get(response, [])
        exception_type = self.exceptions.get(response)
        for i in yielded:
            yield i
        if exception_type is not None:
            raise exception_type


class RedirectUrlResolverTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for RedirectUrlResolver class

    :var valid_urls: a list of strings representing valid urls used
     in tests
    :var head_mock: a mocked implementation of head function
    used by the tested class to perform HEAD requests. Uses an instance
    of HeadSideEffects as its implementation
    :var resolve_redirects_mock: a mock for
    requests.Session.resolve_redirects function. An instance of
    ResolveRedirectsSideEffects is used as its implementation
    :var redirect_results: points to an instance of
    ResolveRedirectsSideEffects used as a replacement implementation
    for requests.Session.resolve_redirects
    :var resolver: an instance of RedirectUrlResolver to be tested
    :var patcher: an object used to patch is_valid_url function
    :var is_valid_url_mock: a mocked implementation of
     the is_valid_url function
    '''
    redirect_url_chain = [
        'http://test.com',
        'http://first.com',
        'http://122.55.33.21',
        'http://[2001:db8:abc:123::42]'
    ]
    no_redirect_url_chain = ['http://noredirects.com']

    def setUp(self):
        session_mock = Mock()
        self.head_mock = session_mock.head
        self.head_mock.side_effect = HeadSideEffects()
        self.resolve_redirects_mock = session_mock.resolve_redirects
        self.redirect_results = ResolveRedirectsSideEffects()
        self.resolve_redirects_mock.side_effect = self.redirect_results
        self.resolver = RedirectUrlResolver(session_mock)
        self.patcher = patch('spam_lists.composites.is_valid_url')
        self.is_valid_url_mock = self.patcher.start()

    def tearDown(self):
        self.patcher.stop()

    def test_get_locations_for_invalid(self):
        self.is_valid_url_mock.return_value = False
        with self.assertRaises(InvalidURLError):
            next(self.resolver.get_locations('http://test.com'))

    def _set_up_side_effects(self, url_histories, exceptions=None,
                             last_location=''):
        ''' Prepare mocks for their calls to have expected side effects

        :param url_histories: a sequence containing sequences of
        url addresses of all responses to a request to a redirecting url
        :param exceptions: a dictionary mapping initial urls of
        redirection chains to exceptions raised when attempting to
        get response to a request to final location
        :param last_location: a value for location response header,
        specifying the location of the last request, if it couldn't be
        completed
        '''
        for history in url_histories:
            response_mocks = [get_response_mock(u) for u in history]
            response_mocks[-1].headers = {'location': last_location}
            first_response = response_mocks.pop(0)
            self.head_mock.side_effect[history[0]] = first_response
            self.redirect_results.responses[first_response] = response_mocks
            if exceptions is None:
                exceptions = {}
            exception_type = exceptions.get(history[0])
            self.redirect_results.exceptions[first_response] = exception_type

    def _test_get_locations(self, argument, expected):
        url_generator = self.resolver.get_locations(argument)
        self.assertEqual(expected, list(url_generator))

    @parameterized.expand([
        ('no_url', no_redirect_url_chain),
        ('urls', redirect_url_chain)
    ])
    def test_get_locations_yields(self, _, history):
        expected = history[1:]
        self._set_up_side_effects([history])
        self._test_get_locations(history[0], expected)

    @parameterized.expand([
        [ConnectionError],
        [InvalidSchema],
        [Timeout],
    ])
    def test_get_locations_arg_raising(self, exception_type):
        self.head_mock.side_effect = exception_type
        self._test_get_locations('http://error_source', [])

    @parameterized.expand([
        ('initial_url_causing_timeout', no_redirect_url_chain, Timeout),
        ('last_url_casuing_timeout', redirect_url_chain, Timeout),
        ('initial_invalid_url', no_redirect_url_chain, InvalidURL, False),
        ('last_invalid_url', redirect_url_chain, InvalidURL, False),
        (
            'initial_url_causing_connection_error',
            no_redirect_url_chain, ConnectionError
        ),
        (
            'last_url_causing_connection_error',
            redirect_url_chain, ConnectionError
        ),
        (
            'initial_invalid_url_causing_connection_error',
            no_redirect_url_chain, ConnectionError,
            False
        ),
        (
            'last_invalid_url_causing_connection_error',
            redirect_url_chain, ConnectionError,
            False
        ),
        ('initial_invalid_schema', no_redirect_url_chain, InvalidSchema),
        ('last_invalid_schema', redirect_url_chain, InvalidSchema),
        (
            'initial_invalid_url_with_invalid_schema',
            no_redirect_url_chain, InvalidSchema,
            False
        ),
        (
            'last_invalid_url_with_invalid_schema',
            redirect_url_chain, InvalidSchema,
            False
        )
    ])
    def test_get_locations_for(self, _, history,
                               exception_type, triggered_by_valid_url=True):
        ''' The get_locations method is expected to yield all
         valid urls appearing as url addresses and location headers in
         response histories for given urls

        :param history: url addresses of all responses in redirection chain
        :param exception_type: a type of exception to be raised while
        getting a response to the last location header value
        :param triggered_by_valid_url: if True, the value of the last
        location header - the one that tiggered an exception - is
        a valid url, and therefore it is also expected to be yielded
        '''
        expected = history[1:]
        exceptions = {history[0]: exception_type}
        error_source = 'http://triggered.error.com'
        self._set_up_side_effects([history], exceptions, error_source)
        if triggered_by_valid_url:
            expected += [error_source]
        else:

            def is_valid_url(url):
                return url in history
            self.is_valid_url_mock.side_effect = is_valid_url
        self._test_get_locations(history[0], expected)

    def _test_get_new_locations(self, histories):
        self._set_up_side_effects(histories)
        input_data = [h[0] for h in histories]
        redirects = [u for h in histories for u in h[1:]]
        expected = list(set(redirects) - set(input_data))
        actual = list(self.resolver.get_new_locations(input_data))
        self.assertCountEqual(expected, actual)

    def test_get_new_locations(self):
        ''' The method is expected to yield only new urls,
        that is urls that were not part of the original input '''
        no_redirects = 'http://noredirects.com'
        histories = [
            [no_redirects],
            ['http://abc.com', no_redirects],
            [
                'http://first.com',
                'http://second.com',
                'http://third.com'
            ]
        ]
        self._test_get_new_locations(histories)

    def test_get_new_unique_locations(self):
        ''' The generator returned by get_new_locations is expected
        to yield no urls that it yielded previously '''
        duplicated_part = [
            'http://first.com',
            'http://second.com',
            'http://third.com'
        ]
        histories = [
            duplicated_part,
            ['http://abc.com'] + duplicated_part
        ]
        self._test_get_new_locations(histories)

    @patch('spam_lists.composites.CachedIterable')
    def test_get_urls_and_locations(self, cached_iterable_mock):
        ''' The method get_urls_and_locations is expected to return
        an instance of CachedIterable.
        '''
        expected = Mock()
        cached_iterable_mock.return_value = expected
        actual = self.resolver.get_urls_and_locations(['http://test.com'])
        self.assertEqual(expected, actual)


@lru_cache()
def get_url_tester_mock(identifier):
    source = Mock()
    source.identifier = identifier
    return source


class UrlTesterChainTest(
        UrlTesterTestBaseMixin,
        TestFunctionDoesNotHandleMixin,
        unittest.TestCase
):
    # pylint: disable=too-many-public-methods
    ''' Tests for UrlTesterChain class

    This class uses get_url_tester_mock function to populate list of
    url testers used by the tested instance

    :var classification: a set of classifications used in tests
    :var tested_instance: an instance of tested class
    '''
    classification = set(['TEST'])
    url_to_source_id = {
        'http://55.44.21.12': ['source_1', 'source_2'],
        'http://test.com': ['source_3'],
        'https://abc.com': ['source_1'],
        'http://[2001:ddd:ccc:111::22]': ['source_1', 'source_2'],
        'http://[2001:abc:111:22::33]': ['source_3']
    }

    def setUp(self):
        url_testers = []
        for _ in range(3):
            tester = Mock()
            tester.any_match.return_value = False
            tester.lookup_matching.return_value = []
            tester.filter_matching.return_value = []
            url_testers.append(tester)
        self.tested_instance = UrlTesterChain(*url_testers)

    def _add_url_tester(self, source_id, matching_urls):
        ''' Add a preconfigured url tester mock to the tested instance

        :param source_id: an identifier for a mocked url tester
        :param matching_urls: a list of urls expected to be matched by
        a service represented by the mocked url tester
        '''
        tester = get_url_tester_mock(source_id)

        def any_match(url):
            return not set(url).isdisjoint(set(matching_urls))
        tester.any_match.side_effect = any_match
        tester.filter_matching.return_value = list(matching_urls)
        url_items = [self._get_item(u, source_id) for u in matching_urls]
        tester.lookup_matching.return_value = url_items
        if tester not in self.tested_instance.url_testers:
            self.tested_instance.url_testers.append(tester)

    def _get_item(self, url, source_id):
        return AddressListItem(
            url,
            get_url_tester_mock(source_id),
            self.classification
        )

    def _get_expected_items_for_urls(self, urls):
        return [self._get_item(u, i) for u, ids
                in list(urls.items()) for i in ids]

    def _set_matching_urls(self, urls):
        ''' Set urls expected to be matched during a test

        The method groups given urls by their source ids: identifiers
         of services expected to report urls associated with them as
         matching. Then, mocks representing url testers are added
         to the tested instance of UrlTesterChain. They are shuffled
         to ensure some of mocked services reporting a match
         will be queried before some that do not.

        :param urls: a dictionary containing
        '''
        by_source_id = defaultdict(list)
        for url, ids in list(urls.items()):
            for i in ids:
                by_source_id[i].append(url)
        for i, urls in list(by_source_id.items()):
            self._add_url_tester(i, urls)
        shuffle(self.tested_instance.url_testers)

    def test_any_match_expecting_true(self):
        self._test_any_match_for(self.url_to_source_id)

    @parameterized.expand([
        ('no_matching_url', {}),
        ('matching_urls', url_to_source_id)
    ])
    def test_lookup_matching_for(self, _, matching_urls):
        self._test_lookup_matching_for(matching_urls)

    @parameterized.expand([
        ('no_matching_url', {}),
        ('matching_urls', url_to_source_id)
    ])
    def test_filter_matching_for(self, _, matching_urls):
        self._test_filter_matching_for(matching_urls)

    @parameterized.expand([
        ('any_match_raises_value_error', 'any_match', ValueError),
        ('any_match_raises_unknown_code_error', 'any_match', UnknownCodeError),
        ('lookup_matching_raises_value_error', 'lookup_matching', ValueError),
        (
            'lookup_matching_raises_unknown_code_error',
            'lookup_matching',
            UnknownCodeError
        ),
        (
            'filter_matching_raises_value_error',
            'filter_matching',
            ValueError
        ),
        (
            'filter_matching_raises_unknown_code_error',
            'filter_matching',
            UnknownCodeError
        )
    ])
    def test_(self, _, function_name, error_type):
        function = getattr(self.tested_instance, function_name)
        for tester in reversed(self.tested_instance.url_testers):
            error_source = getattr(tester, function_name)
            self._test_function_does_not_handle(
                error_type,
                error_source,
                function,
                ['http://triggeringerror.com']
            )


class CachedIterableTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for CachedIterable class

    :var iterator_mock: a mock object representing iterator injected
    into the tested instance
    :var cache: a list of values set as initial cache for tested instance
    :var tested_instance: an instance of CachedIterable to be tested
    '''
    def setUp(self):
        self.iterator_mock = MagicMock()
        self.cache = range(3)
        self.tested_instance = CachedIterable(self.iterator_mock, self.cache)

    def test_cached_returned_first(self):
        for i in self.tested_instance:
            if i == self.cache[-1]:
                break
        self.iterator_mock.__iter__.assert_not_called()

    def test_fixed_order(self):
        self.iterator_mock.__iter__.return_value = range(4, 10)
        first_run_result = list(self.tested_instance)[:9]
        second_run_result = list(self.tested_instance)[:9]
        self.assertSequenceEqual(first_run_result, second_run_result)


class GeneralizedUrlTesterTest(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    ''' Tests for GeneralizedUrlTester class

    :var no_resolution_setup: a parameter setup for test methods
    requiring no explicit parameter to use (or not use)
    redirect resolution in tested calls. Contains names of
    method of tested instance to be called
    :var common_setup: a parameter setup for test methods
    testing calls with and without redirect resolution, depending
    on value of a parameter, with False assumed as default.
    Contains names of method of tested instance to be called.
    :var test_urls: a list of urls passed as argument to methods of
    tested_instance
    :var tested_instance: instance of GeneralizedUrlTester to be
    tested.
    :var whitelist_mock: an object representing an instance of
    whitelist used by tested_instance
    :var url_tester_mock: an object representing a url tester instance
    to be used by tested instance
    :var resolver_mock: an object representing an instance of
    redirect resolver to be used by tested instance
    '''
    test_urls = ['http:abc.com', 'http://def.com', 'http://xyz.com']
    no_resolution_setup = [
        ['any_match'],
        ['filter_matching'],
        ['lookup_matching'],
    ]
    common_setup = [
        ['any_match', True],
        ['filter_matching', True],
        ['lookup_matching', True]
    ] + no_resolution_setup

    def setUp(self):
        self.whitelist_mock = Mock()
        self.whitelist_mock.filter_matching.return_value = MagicMock()
        self.resolver_mock = Mock()
        self.resolver_mock.get_urls_and_locations.return_value = MagicMock()
        self.url_tester_mock = Mock()
        self.tested_instance = GeneralizedUrlTester(
            self.url_tester_mock,
            self.whitelist_mock,
            self.resolver_mock
        )

    def _call_for(self, function_name, resolve_redirects):
        function = getattr(self.tested_instance, function_name)
        return function(self.test_urls, resolve_redirects)

    @parameterized.expand(common_setup)
    def test_whitelist_used_with(self, function_name, resolve_redirects=False):
        self._call_for(function_name, resolve_redirects)
        urls_and_locations = self.test_urls
        if resolve_redirects:
            urls_and_locations = self.resolver_mock.get_urls_and_locations()
        whitelist_method = self.whitelist_mock.filter_matching
        whitelist_method.assert_called_once_with(
            urls_and_locations
        )

    @parameterized.expand(no_resolution_setup)
    def test_resolution_with(self, function_name):
        ''' Redirect resolution must be performed during all calls
        that receive resolve_redirects=True as argument '''
        self._call_for(function_name, True)
        resolver_function = self.resolver_mock.get_urls_and_locations
        resolver_function.assert_called_once_with(self.test_urls)

    @parameterized.expand(no_resolution_setup)
    def test_no_resolution_with(self, function_name):
        ''' Redirect resolution must not be performed during all calls
        that receive resolve_redirects=True as argument '''
        self._call_for(function_name, False)
        self.resolver_mock.assert_not_called()

    @parameterized.expand(common_setup)
    def test_url_tester_results_for(self, function_name,
                                    resolve_redirects=False):
        ''' Each method must return result of a method of url_tester
        called during its execution'''
        url_tester_function = getattr(self.url_tester_mock, function_name)
        expected = url_tester_function()
        actual = self._call_for(function_name, resolve_redirects)
        self.assertEqual(expected, actual)


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
'''
This module contains definitions used by some of the test_ modules
'''
from __future__ import unicode_literals

from builtins import object  # pylint: disable=redefined-builtin
from future.moves.urllib.parse import urlparse
from nose_parameterized import parameterized
from types import GeneratorType

from spam_lists.exceptions import InvalidURLError, InvalidHostError
from spam_lists.structures import AddressListItem
from test.compat import MagicMock, patch, lru_cache


class UrlTesterTestBaseMixin(object):
    ''' A class providing basic methods for performing tests for classes
    having any_match, filter_matching and lookup_matching methods

    The methods are to be used to write and generate actual test methods
    '''
    valid_urls = ['http://test.com', 'http://127.33.22.11',
                  'https://[2001:ddd:ccc:123::55]']

    def _test_any_match_for(self, matching_urls):
        self._set_matching_urls(matching_urls)
        self.assertTrue(
            self.tested_instance.any_match(
                self.valid_urls + list(matching_urls)
            )
        )

    def _test_any_match_returns_false(self, not_matching_urls):
        self.assertFalse(self.tested_instance.any_match(not_matching_urls))

    def _test_lookup_matching_for(self, matching_urls):
        self._set_matching_urls(matching_urls)
        expected = self._get_expected_items_for_urls(matching_urls)
        actual = list(
            self.tested_instance.lookup_matching(
                self.valid_urls + list(matching_urls)
            )
        )

        self.assertCountEqual(expected, actual)

    def _test_filter_matching_for(self, matching_urls):
        self._set_matching_urls(matching_urls)
        actual = list(
            self.tested_instance.filter_matching(
                self.valid_urls + list(matching_urls)
            )
        )

        self.assertCountEqual(matching_urls, actual)

    def test_any_match_returns_false(self):
        self._test_any_match_returns_false(self.valid_urls)


class UrlTesterTestMixin(UrlTesterTestBaseMixin):
    ''' A class providing pre-generated tests for classes
    having any_match, filter_matching and lookup_matching
    methods '''
    classification = set(['TEST'])
    valid_url_input = [
        ('ipv4_url', ['http://55.44.33.21']),
        ('hostname_url', ['https://abc.com']),
        ('ipv6_url', ['http://[2001:ddd:ccc:111::33]'])
    ]
    valid_url_list_input = [
        ('no_matching_url', []),
        ('two_urls', [
            'http://55.44.33.21',
            'https://abc.com'
        ])
    ]+valid_url_input

    @parameterized.expand([
        ('any_match'),
        ('lookup_matching'),
        ('filter_matching')
    ])
    @patch('spam_lists.validation.is_valid_url')
    def test_query_for_invalid_url_with(
            self,
            function_name,
            is_valid_url_mock
    ):
        invalid_url = 'http://invalid.url.com'
        is_valid_url_mock.side_effect = lambda u: u != invalid_url
        function = getattr(self.tested_instance, function_name)
        with self.assertRaises(InvalidURLError):
            function(self.valid_urls + [invalid_url])

    @parameterized.expand(valid_url_input)
    def test_any_match_returns_true_for(self, _, matching_urls):
        self._test_any_match_for(matching_urls)

    @parameterized.expand(valid_url_list_input)
    def test_lookup_matching_for(self, _, matching_urls):
        self._test_lookup_matching_for(matching_urls)

    @parameterized.expand(valid_url_list_input)
    def test_filter_matching_for(self, _, matching_urls):
        self._test_filter_matching_for(matching_urls)

    def _get_expected_items(self, values):
        def get_item(item):
            return AddressListItem(
                item,
                self.tested_instance,
                self.classification
            )
        return [get_item(v) for v in values]


def get_hosts(urls):
    return [urlparse(u).hostname for u in urls]


class HostListTestMixin(UrlTesterTestMixin):
    ''' A common test case for all classes that represent
    a host list stored locally or by a remote service '''
    valid_host_input = [
        ('ipv4', '255.0.120.1'),
        ('hostname', 'test.pl'),
        ('ipv6', '2001:ddd:ccc:111::33')
    ]

    def _set_matching_urls(self, urls):
        self._set_matching_hosts(get_hosts(urls))

    def _get_expected_items_for_urls(self, urls):
        return self._get_expected_items(get_hosts(urls))

    def _get_result_for_invalid_host(self, function):
        unsupported_host = 'unsupported.com'
        self.host_factory_mock.side_effect = InvalidHostError
        return function(unsupported_host)

    def test_contains_for_invalid_host(self):
        function = self.tested_instance.__contains__
        actual = self._get_result_for_invalid_host(function)
        self.assertFalse(actual)

    def test_lookup_for_invalid_host(self):
        function = self.tested_instance.lookup
        actual = self._get_result_for_invalid_host(function)
        self.assertIsNone(actual)

    @parameterized.expand([
        ('__contains__'),
        ('lookup')
    ])
    @patch('spam_lists.validation.is_valid_host')
    def test_invalid_host_query_using(
            self,
            function_name,
            is_valid_host_mock
    ):
        function = getattr(self.tested_instance, function_name)
        invalid_host = 'invalid.com'
        is_valid_host_mock.side_effect = lambda h: h != invalid_host
        with self.assertRaises(InvalidHostError):
            function(invalid_host)

    @parameterized.expand(valid_host_input)
    def test_contains_for_listed(self, _, value):
        self._set_matching_hosts([value])
        self.assertTrue(value in self.tested_instance)

    @parameterized.expand(valid_host_input)
    def test_contains_for_not_listed(self, _, value):
        self.assertFalse(value in self.tested_instance)

    @parameterized.expand(valid_host_input)
    def test_lookup_for_listed(self, _, value):
        expected = self._get_expected_items([value])[0]
        self._set_matching_hosts([value])
        self.assertEqual(self.tested_instance.lookup(value), expected)

    @parameterized.expand(valid_host_input)
    def test_lookup_for_not_listed(self, _, value):
        self.assertIsNone(self.tested_instance.lookup(value))


@lru_cache()
def host_list_host_factory(host):
    host_object = MagicMock()
    host_object.to_unicode.return_value = host
    return host_object


class TestFunctionDoesNotHandleMixin(object):

    def _test_function_does_not_handle(self, exception_type, exception_origin,
                                       function, *args, **kwargs):
        '''
        Test if a given function does not handle an error
        raised by a dependency

        :param exception_type: a type of exception to be raised
        :param exception_origin: a function raising the error,
        represented by an instance of Mock
        :param function: a function to be tested
        :param *args: positional arguments to be passed to
        the tested function
        :param **kwargs: keyword arguments to be passed to
        the tested function
        '''
        exception_origin.side_effect = exception_type
        with self.assertRaises(exception_type):
            result = function(*args, **kwargs)
            if isinstance(result, GeneratorType):
                list(result)



# -*- coding: utf-8 -*-

'''
This module defines classes of objects containing various data, and
functions for creating proper objects for given arguments. They are used
by service models in spam_lists.host_list module
'''

from __future__ import unicode_literals

from collections import namedtuple
import ipaddress

from builtins import str, object  # pylint: disable=redefined-builtin
from dns import name
from dns.reversename import (
    ipv4_reverse_domain, ipv6_reverse_domain, from_address as name_from_ip
)
from future.utils import raise_with_traceback
import tldextract
import validators

from .exceptions import (
    InvalidHostError, InvalidHostnameError, InvalidIPv4Error, InvalidIPv6Error
)
from .compat import lru_cache


class Host(object):
    ''' Base class for host objects '''
    def __lt__(self, other):
        ''' Check if the other is smaller

        This method is necessary for sorting and search
        algorithms using bisect_right.

        :param other: a value to be compared
        :returns: result of comparison between value attributes of
        both this object and the other, or of comparison between
        their unicode string representations.

        In case of the other not having necessary attributes,
        NotImplemented constant is returned.
        '''
        try:
            try:
                return self.value < other.value
            except TypeError:
                return self.to_unicode() < other.to_unicode()
        except AttributeError:
            return NotImplemented


class Hostname(Host):
    ''' A class of objects representing hostname values.

    The instances are used as values tested by clients of
    hostname-listing services or as items stored by objects
    representing such host lists.
    '''
    def __init__(self, value):
        ''' Create a new instance of Hostname

        :param value: a string representing a hostname
        :raises InvalidHostnameError: if value parameter is not a valid domain
        '''
        value = str(value)
        if not validators.domain(value):
            msg = "'{}' is not a valid hostname".format(value)
            raise_with_traceback(InvalidHostnameError(msg))
        hostname = name.Name(value.split('.'))
        self.value = hostname
        self.relative_domain = hostname

    def is_subdomain(self, other):
        ''' Test if the object is a subdomain of the
        other

        :param other: the object to which we compare this instance
        :returns: True if this instance is a subdomain of the other
        '''
        compared = other.value if hasattr(other, 'value') else other
        try:
            return self.value.is_subdomain(compared)
        except AttributeError:
            return False

    is_match = is_subdomain

    def to_unicode(self):
        ''' Get unicode string representing the object

        :returns: the ip value as unicode string
        '''
        return self.value.to_unicode()


class IPAddress(Host):
    ''' A class of objects representing IP address values.

    The instances are used as values tested by clients of
    IP-address-listing services or as items stored by objects
    representing such IP address lists.
    '''
    reverse_domain = None

    def __init__(self, value):
        ''' Constructor

        :param value: a valid ip address for this class
        :raises self.invalid_ip_error_type: if the value is not
        a valid ip address for this class
        '''
        try:
            self.value = self.factory(value)
        except ValueError:
            msg_tpl = '{} is not a valid ip address for {}'
            msg = msg_tpl.format(value, self.__class__)
            raise_with_traceback(self.invalid_ip_error_type(msg))

    @property
    def relative_domain(self):
        ''' Get a relative domain name representing the ip address

        :returns: the reverse pointer relative to the common root
        depending on the version of ip address represented by this object
        '''

        return name_from_ip(str(self.value)).relativize(self.reverse_domain)

    def is_subdomain(self, _):
        # pylint: disable=no-self-use
        ''' Check if this object is a subdomain of the other

        :param other: another host
        :returns: False, because ip address is not a domain
        '''
        return False

    def to_unicode(self):
        ''' Get unicode string representing the object

        :returns: the ip value as unicode string
        '''
        return str(self.value)

    def is_match(self, other):
        return self == other


class IPv4Address(IPAddress):
    factory = ipaddress.IPv4Address
    reverse_domain = ipv4_reverse_domain
    invalid_ip_error_type = InvalidIPv4Error


class IPv6Address(IPAddress):
    factory = ipaddress.IPv6Address
    reverse_domain = ipv6_reverse_domain
    invalid_ip_error_type = InvalidIPv6Error


def cached(function):
    return lru_cache()(function)

hostname = cached(Hostname)
ip_v4 = cached(IPv4Address)
ip_v6 = cached(IPv6Address)


@cached
def create_host(factories, value):
    ''' Create an instance of host object for given value, using
    the factories.

    :param factories: a list of functions that return host objects
    (Hostname, IPv4Address, IPv6Address) for valid arguments
    :param value: a value to be passed as argument to factories
    :returns: an object representing value, created by one of the factories.
    It's a return value of the first factory that could create it
    for the given argument
    :raises InvalidHostError: if the value is not a valid input
    for any factory used by this function
    '''
    data = [value]
    for func in factories:
        try:
            return func(value)
        except InvalidHostError as ex:
            data.append(str(ex))
    msg_tpl = "Failed to create a host object for '{}', \
    raising the following errors in the process:"+"\n".join(data)
    raise InvalidHostError(msg_tpl.format(value))


def ip_address(value):
    ''' Create an ip address object

    :param value: a valid ip address
    :returns: a .structures.IPAddress subclass instance
    :raises InvalidHostError: if the value is not a valid IPv4 or
    IPv6 value
    '''
    factories = ip_v4, ip_v6
    return create_host(factories, value)


def hostname_or_ip(value):
    ''' Create a hostname or ip address object
    for given value

    :param value: a valid host string
    :returns: a host object for given value
    :raises InvalidHostError: if the value is not a valid hostname or
    ip address
    '''
    factories = ip_v4, ip_v6, hostname
    return create_host(factories, value)


TLD_EXTRACTOR = tldextract.TLDExtract()


def registered_domain(value):
    ''' Create a Hostname instance representing registered domain
    extracted from the value

    :param value: a valid host string
    :returns: a Hostname instance representing registered domain
    :raises InvalidHostnameError: if the value is not a valid hostname
    '''
    registered_domain_string = TLD_EXTRACTOR(value).registered_domain
    return hostname(registered_domain_string)


def registered_domain_or_ip(value):
    ''' Get host object representing a registered domain or an ip address

    :param value: a valid hostname or ip string
    :returns: a host object representing a registered domain extracted from
    given hostname, or an ip address
    :raises InvalidHostError: if value is not a valid host
    '''
    factories = ip_v4, ip_v6, registered_domain
    return create_host(factories, value)


def non_ipv6_host(value):
    ''' Create host object representing a registered domain or an IPv4 address

    :param value: a valid hostname or IPv4 string
    :returns: a host object representing a registered domain extracted from
    given hostname, or an IPv4 address
    :raises InvalidHostError: if value is not a valid hostname or IPv4 address
    '''
    factories = ip_v4, registered_domain
    return create_host(factories, value)

AddressListItem = namedtuple('AddressListItem', 'value source classification')



# -*- coding: utf-8 -*-
from codecs import open
import re
from setuptools import setup
import sys


def read_attributes(string, *names):
    regex_tpl = r'^__{}__\s*=\s*[\'"]([^\'"]*)[\'"]'

    def read(name):
        regex = regex_tpl.format(name)
        return re.search(regex, string, re.MULTILINE).group(1)

    return [read(n) for n in names]


with open('spam_lists/__init__.py', 'r') as fd:
    content = fd.read()
    name, version, author, _license = read_attributes(
        content,
        'title',
        'version',
        'author',
        'license'
    )

with open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()


install_requires = ['future', 'requests', 'tldextract', 'validators']
tests_require = ['nose-parameterized']

if sys.version_info.major < 3:
    install_requires += ['cachetools', 'dnspython', 'ipaddress']
    tests_require += ['mock']

    if sys.version_info < (2, 7, 9):
        ''' request[security] extras '''
        install_requires += ['pyOpenSSL>=0.13', 'ndg-httpsclient', 'pyasn1']
else:
    install_requires += ['dnspython3']

setup(
    name=name,
    version=version,
    description='Web address blacklist/whitelist library for Python',
    long_description=readme,
    author=author,
    author_email='piotr.rusin88@gmail.com',
    url='https://github.com/piotr-rusin/spam-lists',
    packages=['spam_lists', 'test', 'test.integration', 'test.unit'],
    install_requires=install_requires,
    license=_license,
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
    ),
    keywords=('spam dnsbl surbl google-safe-browsing-api '
              'spamhaus whitelist blacklist'),
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
        'dev-tools': ['pylint', 'restview']
    },
)
